(* ::Package:: *)

(*AtomicDensityMatrix*)

(*By Simon Rochester*)
(*U.C. Berkeley Physics Dept.*)

(*The AtomicDensityMatrix package provides an environment for performing atomic physics calculations.*)

(*\[Copyright] Copyright 2004-2009, Simon Rochester*)

(*Mathematica version*)
(*6.0+*)


(*History*)
(*First version set up as a package 2005-2-21.*)
(*See about.html for version history.*)

(*Unprotect WignerD if defined*)
If[$VersionNumber >= 8., Unprotect[WignerD];WignerD[{1,1,1},0],If[DownValues[WignerD]=={},Remove[WignerD]]]

BeginPackage["AtomicDensityMatrix`AtomicDensityMatrix`",{
  (*"PhysicalConstants`","Units`",*)
  "AtomicDensityMatrix`UsageMessages`",
  "AtomicDensityMatrix`Common`",
  "AtomicDensityMatrix`CouplingCoefficients`",
  "AtomicDensityMatrix`TensorAlgebra`",
  "AtomicDensityMatrix`AtomicSystem`",
  "AtomicDensityMatrix`ExperimentalSystem`",
  "AtomicDensityMatrix`DensityMatrix`",
  "AtomicDensityMatrix`WignerEckart`", 
  "AtomicDensityMatrix`WignerRotate`"}]

Begin["`Private`"]


(* ::Subsection:: *)
(*Meta info*)


$ADMVersion="13.07.19"


(* ::Subsection:: *)
(*Tensor decomposition*)


Options[PolarizationMoments]=DeleteCases[Options[DensityMatrix]/.(r_->_)->(r->Automatic),Representation->_]

PolarizationMoments[j_?HalfIntegerQ,opts___Rule] := PolarizationMoments[ZeemanSystem[j], opts]

PolarizationMoments[syst:{_AtomicState..},opts___Rule] :=
	Module[{newopts, rfunc},
		newopts = 
			Join[
	   			DeleteCases[Join[{opts},Options[PolarizationMoments]],_->Automatic],
	   			Options[DensityMatrix]
	   		]/.(DMLabel->(None|False))->(DMLabel->Sequence[]);
	    rfunc[l1_,l2_,k_,q_] = 
	    	(Subscript[DMSymbol, DMLabel,{l1,l2},{k,q}][TimeVariable]/.newopts)[[If[TimeDependence/.newopts, All, 0]]];
	    DirectSelfProduct[
	    	Table[
	      		Transpose@{Table[rfunc[#1[[1]],#2[[1]],k,q],{q,k,-k,-1}]},
	      		{k,Abs[#1[[2]]-#2[[2]]],#1[[2]]+#2[[2]]}
	      	]&,
	      	{LevelLabel[#],F[#]}&/@RemoveZeemanSublevels[syst]
	    ]
	]

PolarizationMatrices[j_?HalfIntegerQ] := Component@PolarizationMatrices[ZeemanSystem[j]]

PolarizationMatrices[syst:{_AtomicState..}] :=
   Module[{r},
      ECT[PolarizationMoments[syst, TimeDependence -> False, DMSymbol->r]] /. 
      	Subscript[r, {s1_, s2_}, {k_, q_}] :> WignerEckart[syst, {Polarization[s1, s2], k, q}]
   ]

(*Convert between m m' and \[Kappa] q representation (contravariant components) using the polarization operator. [Var88, Sec. 6.1.5, Eqs. (49,50)], also [Ale2004, Eqs. (A3-8) ]*)

Decomposition[syst:{_AtomicState..},X_?OperatorQ]:=
	Map[Tr[#\[CenterDot]X]&,ECT@PolarizationMatrices[syst],{5}]

Decomposition[N_?OperatorQ]:=Component@Decomposition[ZeemanSystem[Rank[N]],N]

Recomposition[X_?TensorDecompositionQ]:=Recomposition[ZeemanSystem[Rank[X]],{{X}}]

Recomposition[syst:{_AtomicState..},X_?ContravariantDecompositionMatrixQ]:=PolarizationMatrices[syst]\[CenterDot]X

Recomposition[syst:{_AtomicState..},X_?CovariantDecompositionMatrixQ]:=X\[CenterDot]ECT@PolarizationMatrices[syst]

Recomposition[J_?HalfIntegerQ,X_?ContravariantQ]:=Component[PolarizationMatrices[J],Rank[X]]\[CenterDot]X

Recomposition[J_?HalfIntegerQ,X_?CovariantQ]:=X\[CenterDot]ECT@Component[PolarizationMatrices[J],Rank[X]]



(* ::Subsection:: *)
(*Angular momentum probability distribution*)


(*Angular momentum probability distribution in terms of the \[Kappa]-q tensor elements [Ale2004, Eq. (A11)]*)

\[Rho]JJ=MaxProjectionProbability;

MaxProjectionProbability[T_?TensorDecompositionQ,th_,ph_]:=Module[
	{J=Rank[T],T1=ToContravariant@T},
	Sqrt[(4 Pi)/(2J+1)]
	*Sum[
		CG[{J,J},{k,0},{J,J}]Component[T1,k,q]SphericalHarmonicY[k,q,th,ph],
		{k,0,2J},{q,-k,k}
	]
]

(*Angular momentum probability distribution in terms of the m m' tensor elements [Ale2004, Eq. (A12)]*)

MaxProjectionProbability[\[Rho]_?OperatorQ,\[Theta]_,ph_]:=
	(ECT[#]\[CenterDot]\[Rho]\[CenterDot]#&)[WignerD[Rank[\[Rho]],{ph,\[Theta],0}]\[CenterDot]StretchedState[Rank[\[Rho]]]]

Options[AMPSPlot]=Options[SphericalPlot3D]

AMPSPlot[\[Rho]_,opts___Rule]:=
	SphericalPlot3D[Evaluate[Re[MaxProjectionProbability[\[Rho],\[Theta],ph]]],{\[Theta],0,\[Pi]},{ph,0,2\[Pi]},opts,Evaluate[Sequence@@Options[AMPSPlot]]]


(* ::Subsection:: *)
(*Optical field*)


Options[OpticalField] = {PolarizationVector -> {1, 0, 0}, 
  PropagationVector -> {0, 0, 1}, 
  Parameterization -> "AngleEllipticity", 
  CartesianCoordinates -> {Xx, Yy, Zz}, TimeVariable -> Automatic}
  
(*Takes the arguments and options to OpticalField and puts them in standardized form.*)

StandardizeField[args__, opts___Rule] /; 2 <= Length[{args}] <= 3 := 
 Module[{args1 = EnsureList /@ {args}, 
   defpol = 
    Switch[Parameterization /. {opts} /. Options[OpticalField], 
     "AngleEllipticity", {0, 0}, "RatioPhase", {0, 0}, "Stokes", {1, 0, 0}, 
     "SphericalPhase", {0, 0}]}, {Join[args1[[1]], 
    Drop[{0, 0}, Length[args1[[1]]]]], 
   Join[args1[[2]], Drop[{0, 0}, Length[args1[[2]]]]], 
   If[Length[args1] == 3, 
    Join[args1[[3]], Drop[defpol, Length[args1[[3]]]]], defpol]}]

OpticalField[args__, opts : OptionsPattern[]] /; 
  2 <= Length[{args}] <= 3 := 
 Module[{w, k0, E0, e, b, k, r, al, ep, \[Chi], cphi, ph, S1, 
   S2, S3, f, pol, t, Ep, Em, php, 
   phm,v0}, {{w, k0}, {E0, cphi}, pol} = 
   StandardizeField[args, opts];
  t = OptionValue[TimeVariable] /. Automatic -> TimeVariable /. 
    Options[DensityMatrix];
  {e, k} = {Normalize@OptionValue[PolarizationVector], 
    Normalize@OptionValue[PropagationVector]};
  r = CartesianCoordinates /. {opts} /. Options[OpticalField];
  b = Cross[k, e];
  f = E0 E^(I (k0 k.r + cphi - t w));
  Switch[OptionValue[Parameterization],
   "AngleEllipticity", {al, ep} = pol; 
   f RotationMatrix[al, k].(Cos[ep] e + I Sin[ep] b), 
   "RatioPhase", {\[Chi], ph} = pol; 
   f (Cos[\[Chi]] e + Sin[\[Chi]] E^(I ph) b),
   "Stokes", {S1, S2, S3} = pol; 
   f (b E^(I If[S2 === S3 === 0, \[Pi]/4, ArcTan[S2, S3]]) Sqrt[
        1/2 - S1/2] + (e Sqrt[1 + S1])/Sqrt[2]),
   "SphericalPhase", {Ep, php} = {E0, cphi}; {Em, phm} = pol;
   v0 = ToCartesian[{{E^(I php) Ep}, {0}, {E^(I phm) Em}}];
    E^(I (k0 k.r - t w)) (v0[[1]] e + v0[[2]] b)
   ]
  ]

ReparameterizeField[{S1_,S2_,S3_},"Stokes"->"RatioPhase"]:={ArcCos[Sqrt[1+S1]/Sqrt[2]],If[N@S2===N@S3===0.,\[Pi]/4,ArcTan[S2,S3]]}

ReparameterizeField[{\[Chi]_,ph_},"RatioPhase"->"Stokes"]:={Cos[2 \[Chi]],Cos[ph] Sin[2\[Chi]],Sin[ph] Sin[2 \[Chi]]}

ReparameterizeField[{al_,ep_},"AngleEllipticity"->"Stokes"]:={Cos[2 al] Cos[2 ep],Cos[2 ep] Sin[2 al],Sin[2 ep]}

ReparameterizeField[{S1_,S2_,S3_},"Stokes"->"AngleEllipticity"]:={If[N@S1===N@S2===0.,0,(Sign[S2]+KroneckerDelta[S2])ArcSin[Sqrt[1/2-S1/(2 Sqrt[S1^2+S2^2+.1])]]],(Sign[S3]+KroneckerDelta[S3])ArcCos[Sqrt[1/2+1/2 Sqrt[S1^2+S2^2]]]}

ReparameterizeField[{\[Chi]_,ph_},"RatioPhase"->"AngleEllipticity"]:=ReparameterizeField[ReparameterizeField[{\[Chi],ph},"RatioPhase"->"Stokes"],"Stokes"->"AngleEllipticity"]

ReparameterizeField[{al_,ep_},"AngleEllipticity"->"RatioPhase"]:=ReparameterizeField[ReparameterizeField[{al,ep},"AngleEllipticity"->"Stokes"],"Stokes"-> "RatioPhase"]



(* ::Subsection:: *)
(*Hamiltonians*)


(* ::Subsubsection:: *)
(*Unperturbed Hamiltonian*)


InternalHamiltonian[system:{_AtomicState..},opts___Rule]:=WignerEckart[system,{Energy,0,0},opts]

Energies[system:{_AtomicState..},crit_]:=Union[WignerEckart[#,{Energy,0,0},#]&/@SelectStates[system,crit]]

TransitionFrequencies[system:{_AtomicState..},label1_,label2_]:=
	Transpose@Expand@Outer[Plus,Energies[system,label2],-Energies[system,label1]]



(* ::Subsubsection:: *)
(*E1 Hamiltonian*)


(*Subscript[H, l]=-d\[CenterDot]E*)

Options[E1Hamiltonian]={AtomicSystemType->Automatic};

E1Hamiltonian[system:{_AtomicState..},efield_?CartesianQ,opts___Rule]:=If[
ASTOpt[{opts},E1Hamiltonian,system]===Toy||ScalarQ[efield],
-ComplexExpand[Re[Sqrt[efield\[CenterDot]efield]//PowerExpand]]WignerEckart[system,{Dipole,1,0},opts],
-ComplexExpand[Re[efield]]\[CenterDot]ToCartesian[WignerEckart[system,{Dipole,1},opts]]
]

(*E1Hamiltonian[system:{_AtomicState..},{efield_,label_},opts___Rule]:=E1Hamiltonian[system,efield,Label->label,opts]*)



(* ::Subsubsection:: *)
(*Magnetic field Hamiltonian*)


(*Subscript[H, B]=-\[Mu]\[CenterDot]B*)

MagneticHamiltonian[system:{_AtomicState..},bf_?CartesianQ,opts___Rule]:=-bf\[CenterDot]ToCartesian[WignerEckart[system,{MagneticMoment,1},opts]]

MagneticHamiltonian[system:{_AtomicState..},bf_?ScalarQ,opts___Rule]:=MagneticHamiltonian[system,{0,0,bf},opts]



(* ::Subsubsection:: *)
(*Effective electric field Hamiltonian*)


(*We define the polarizability operator of rank \[Kappa], denoted Subscript[\[Alpha], \[Kappa]], by*)
(*	Subscript[*)(*Subscript[H, Stark]={*)(*StyleBox["E",*)(*FontWeight->"Bold",*)(*FontSlant->"Plain"]\[CircleTimes]*)(*SuperscriptBox[*)(*StyleBox["E",*)(*FontWeight->"Bold",*)(*FontSlant->"Plain"],"*"]}, \[Kappa]]\[CenterDot]Subscript[\[Alpha], \[Kappa]] ,*)
(*where Subscript[H, Stark] is effective Stark Hamiltonian and E is the (complex) electric field. In terms of the polarizabilities, the diagonal elements for a z-directed field are given by*)
(*	Subscript[\[CapitalDelta], Stark]=-(E^2/2)(Subscript[\[Alpha], 0]+Subscript[\[Alpha], 2] (3m^2-j(j+1))/j(2j-1)).*)
(*Comparison of these equations allows the determination of the reduced matrix elements of Subscript[\[Alpha], \[Kappa]].*)

PolarizabilityHamiltonian[system:{_AtomicState..},efield_?CartesianQ,k_,opts___Rule]:=
SphericalTensorProduct[ToCovariant@efield,ToCovariant@EC@efield,{k}]\[CenterDot]ECT@WignerEckart[system,{Polarizability,k},opts]

(*PolarizabilityHamiltonian[system:{_AtomicState..},{efield_?CartesianQ,label_},k_,opts___Rule]:=
PolarizabilityHamiltonian[system,efield,k,Label->label,opts]*)

PolarizabilityHamiltonian[system:{_AtomicState..},efield_?ScalarQ,k_,opts__Rule]:=PolarizabilityHamiltonian[system,{0,0,efield},k,opts]

PolarizabilityHamiltonian[system:{_AtomicState..},efield_,opts___Rule]:=Sum[PolarizabilityHamiltonian[system,efield,k,opts],{k,0,2}]



(* ::Subsubsection:: *)
(*Transition Polarizability Hamiltonian*)


TransitionPolarizabilityHamiltonian[system:{_AtomicState..},opticalfield_?CartesianQ,staticfield_?CartesianQ,k_,opts___Rule]:=
SphericalTensorProduct[ToCovariant@staticfield,ToCovariant@ComplexExpand@Re@opticalfield,{k}]\[CenterDot]ECT@WignerEckart[system,{TransitionPolarizability,k},opts]

TransitionPolarizabilityHamiltonian[system:{_AtomicState..},opticalfield_, staticfield_,opts___Rule]:=Sum[TransitionPolarizabilityHamiltonian[system,opticalfield, staticfield,k,opts],{k,0,2}]



(* ::Subsubsection:: *)
(*Total Hamiltonian*)


Options[Hamiltonian]={ElectricField->0,MagneticField->0,Interaction->Automatic,AtomicSystemType->Automatic}

Hamiltonian[system:{_AtomicState..},expSystem:{_Region..},opts:OptionsPattern[]] := 
	Hamiltonian[system,#,opts]&/@expSystem

Hamiltonian[system:{_AtomicState..},region_Region,opts:OptionsPattern[]] := 
	Hamiltonian[system,
		ElectricField->If[DefinedQ[region,ElectricField],ElectricField[region], {0,0,0}]
			+If[DefinedQ[region,OpticalParameters],OpticalField@@(OpticalParameters[region]), {0,0,0}],
		MagneticField->If[DefinedQ[region,MagneticField],MagneticField[region], {0,0,0}], 
	opts
	]

Hamiltonian[system:{_AtomicState..},opts___Rule] :=
  Module[ {EF,BF,type,zero},
    {EF,BF,type} = {ElectricField,MagneticField,Interaction}/. {opts}/. Options[Hamiltonian];
    If[ type===All,
      type = {Internal,E1,Magnetic,Polarizability,TransitionPolarizability}
    ];
    If[ type===Automatic,
      type = {Internal,E1,Magnetic}
    ];
    If[ Head[type]=!=List,
      type = {type}
    ];
    If[ EF===0,
      EF = {0,0,0}
    ];
    EF = If[ MatchQ[EF,{{_?CartesianQ,_}..}],
           EF,
           {EF}
         ];
    wigopts = Sequence@@({Sequence@@FilterRules[Flatten[{opts}],Options[WignerEckart]]}\[Union]{Sequence@@FilterRules[Flatten[{opts}],Options[ReducedME]]});
    zero = WignerEckart[system,{Zero,0,0}];
    If[ MemberQ[type,Internal],
      InternalHamiltonian[system,wigopts],
      zero
    ]
    + If[ MemberQ[type,E1],
        Total[(E1Hamiltonian[system,#1,wigopts]&)/@EF],
        zero
      ]
    + If[ MemberQ[type,Magnetic]&&ASTOpt[{opts},Hamiltonian,system]=!=Toy,
          MagneticHamiltonian[system,BF,wigopts],
          zero
       ]
    + If[ MemberQ[type,Polarizability],
            Total[(PolarizabilityHamiltonian[system,#1,wigopts]&)/@EF],
            zero
       ]
    + If[ MemberQ[type,TransitionPolarizability],
              TransitionPolarizabilityHamiltonian[system,Cases[EF,{x_,AC}->x,1][[1]],Cases[EF,{x_,DC}->x,1][[1]],wigopts],
              zero
       ]
  ]



(* ::Subsection:: *)
(*Rotating wave approximation*)


(* ::Text::Bold:: *)
(*Should fix this so it doesn't require energy to be set!*)


(*Transformation of the Hamiltonian H under a passive time dependent unitary transformation U. Uses*)
(*H|\[Psi]\[RightAngleBracket]=I d/dt|\[Psi]\[RightAngleBracket] *)
(*and*)
(*|*)
(*Overscript[\[Psi], ~]\[RightAngleBracket]=U^\[Dagger]|\[Psi]\[RightAngleBracket]*)
(*for a passive transformation.*)

EffectiveHamiltonian[H_,U_,t_]:=ECT[U]\[CenterDot]H\[CenterDot]U-I ECT[U]\[CenterDot]D[U,t]

(*Throw out fast-oscillating terms*)

DropFastTerms[expr_, freq_, t_] := 
 Module[{freq1, expr1}, 
 	{expr1, freq1} = Expand[{expr, EnsureList@freq} 
   		//. (f : Sin | Cos)[x_ /; ! FreeQ[x, t]] :> TrigToExp[f[ExpandAll[x]]] 
   		//. Exp[x_ /; ! FreeQ[x, t]] :> Exp[ExpandAll[x]]
   	];
  expr1 /. 
   Flatten[{{Exp[ExpandAll[2 I # t + c_.]] -> 0, 
        Exp[ExpandAll[-2 I # t + c_.]] -> 0} & /@ freq1}]
]
        

(*DropFastTerms[expr_,freq_,t_]:=
	Module[{freq1=EnsureList[freq]},
		Expand[expr /. {
			(f : Sin | Cos)[x_ /; ! FreeQ[x, t]] :> TrigToExp[f[ExpandAll[x]]], 
			Exp[x_] :> Exp[ExpandAll[x]]
		}] /. 
		Flatten[{
			{Exp[ExpandAll[2 I # t + c_.]] -> 0, Exp[ExpandAll[-2 I # t + c_.]] -> 0} & /@ freq1(*,
			If[
				Length[freq1]>1,
				{Exp[ExpandAll[I Total[#]t + c_.]] -> 0, Exp[ExpandAll[-I Total[#]t + c_.]] -> 0} & /@ Subsets[freq1,{2}],
				{}
			]*)
		}]
	]
*)
Options[RotatingWaveTransformMatrix] = {TimeVariable -> Automatic, ReturnFrequencyReplacements -> False, Method -> Automatic}

RotatingWaveTransformMatrix[syst:{_AtomicState..}, 
  freqlist : {{_, _Rule | {__Rule}, ___} ..} | {Except[_Rule], _Rule | {__Rule}, ___}, OptionsPattern[]] := 
 Module[{t, shiftlist, freqreps, freqbackreps, freq, freqlist1, 
   lowerpos, upperpos, lowerShifts, upperShifts, lowerUnfixed, 
   upperUnfixed, Unfixed, transmat, coupling, diffFreq,i,j,k},
  t = OptionValue[TimeVariable] /. Automatic -> TimeVariable /. Options[DensityMatrix];
  (*initialize shiftlist and frequency replacement list*)
  shiftlist = ConstantArray[Unfixed, Length[syst]];
  freqreps = {}; freqbackreps = {};
  (*expand multiple couplings in one entry in freqlist to multiple entries, and delete third elements *)
  freqlist1 = If[MatchQ[freqlist,{_, _Rule | {__Rule}, ___}], {freqlist}, freqlist ];
  freqlist1 = 
   freqlist1[[All, ;; 2]] /. {{f_, c : {__Rule}} :> 
      Sequence @@ ({f, #} & /@ c)};
  (*loop over optical frequencies, defining transform-induced level shifts to eliminate frequencies when possible. 
  If not possible, define replacements to write frequencies in terms of frequencies that have been eliminated.*)
  Do[freq = freqlist1[[i, 1]];
   coupling = freqlist1[[i, 2]];
   (* Find the positions of states specified to be coupled by this frequency *)
   {lowerpos, upperpos} = 
    List @@ (StatePosition[syst, #] & /@ coupling);
   (* Find the previously specified level shifts for these states *)
   {lowerShifts, upperShifts} = 
    Extract[shiftlist, #] & /@ {lowerpos, upperpos};
   (* Split out sublists of states that have different previously specified energy shifts*)
   lowerShifts = Transpose[{lowerShifts, lowerpos}];
   upperShifts = Transpose[{upperShifts, upperpos}];
   {lowerShifts, upperShifts} = 
    Split[#, First[#1]===First[#2]&] & /@ {lowerShifts, upperShifts};
   {lowerShifts, lowerpos} = {lowerShifts[[All, All, 1]], 
     lowerShifts[[All, All, 2]]};
   {upperShifts, upperpos} = {upperShifts[[All, All, 1]], 
     upperShifts[[All, All, 2]]};
   (* Find groups of unfixed states *)
   {lowerUnfixed, upperUnfixed} = 
    Map[AllSameQ[#, Unfixed] &, {lowerShifts, upperShifts}, {2}];
   (* Loop over the sublists of lower and upper states for this frequency, eliminating optical frequencies by shifting states where possible. *)
   Do[
    Which[
     (* If neither upper or lower are fixed, and Method->Automatic or "ShiftUpperState", fix the lower at its current position and shift the upper down *)
     lowerUnfixed[[j]] && upperUnfixed[[k]] && (OptionValue[Method]===Automatic || MemberQ[EnsureList@OptionValue[Method],"ShiftUpperState"]),
     shiftlist = ReplacePart[shiftlist, lowerpos[[j]] -> 0]; 
     shiftlist = ReplacePart[shiftlist, upperpos[[k]] -> -freq],
     (* If neither are fixed, and Method->"ShiftLowerState", fix the upper at its current position and shift the lower up *)
     lowerUnfixed[[j]] && upperUnfixed[[k]] && OptionValue[Method]==="ShiftLowerState",
     shiftlist = ReplacePart[shiftlist, lowerpos[[j]] -> freq]; 
     shiftlist = ReplacePart[shiftlist, upperpos[[k]] -> 0],
     (* If the lower is fixed and the upper is unfixed, 
     shift upper to compensate for lower shift, 
     then shift down by optical frequency *)
     ! lowerUnfixed[[j]] && upperUnfixed[[k]],
     shiftlist = 
      ReplacePart[shiftlist, 
       upperpos[[k]] -> First[lowerShifts[[j]]] - freq],
     (* If the lower is unfixed and the upper is fixed, shift lower to compensate for upper shift, then shift up by optical frequency *)
     lowerUnfixed[[j]] && ! upperUnfixed[[k]],
     shiftlist = 
      ReplacePart[shiftlist, 
       lowerpos[[j]] -> First[upperShifts[[k]]] + freq],
     (* If both are fixed, find replacement rule and back substitution rule for writing the optical frequency in terms of the previously eliminated frequencies and difference frequency *)
     ! lowerUnfixed[[j]] && ! upperUnfixed[[k]],
     diffFreq = Unique[];
     AppendTo[
      freqreps, {freq -> -(First[upperShifts[[k]]] - 
            First[lowerShifts[[j]]]) + diffFreq, {Flatten@
         lowerpos[[j]], Flatten@upperpos[[k]]}}];
     AppendTo[
      freqbackreps, {diffFreq -> 
        freq + (First[upperShifts[[k]]] - 
           First[lowerShifts[[j]]]), {Flatten@lowerpos[[j]], 
        Flatten@upperpos[[k]]}}]
     ], {j, Length[lowerpos]}, {k, Length[upperpos]}],
   {i, Length[freqlist1]}];
  transmat = MatrixExp[I t DiagonalMatrix[shiftlist /. Unfixed -> 0]];
  If[OptionValue[ReturnFrequencyReplacements], {transmat, freqreps, freqbackreps}, transmat]
]
  
RotatingWaveTransformMatrix[system:{_AtomicState..}, 
  shiftlist : {__Rule}, OptionsPattern[]] := Module[{reps, shifts, t},
  t = OptionValue[TimeVariable] /. Automatic -> TimeVariable /. Options[DensityMatrix];
  reps = (Alternatives @@ SelectStates[system, #[[1]]] -> #[[2]] &) /@
     shiftlist;
  shifts = system /. reps /. AtomicState[__] -> 0;
  MatrixExp[I t DiagonalMatrix[shifts]]
]

RotatingWaveTransformMatrix[syst:{_AtomicState..}, Except[_Rule, f_], 
  OptionsPattern[]] := Module[{t},
  t = OptionValue[TimeVariable] /. Automatic -> TimeVariable /. Options[DensityMatrix]; 
  MatrixExp[-I t Chop@
     Replace[DiagonalMatrix[Energy /@ syst] /. 0. -> 0, 
      x_ /; x =!= 0 -> f, {2}]]
]

(*RotatingWaveTransformMatrix[syst:{_AtomicState..},f_]:=MatrixExp[-I t Chop@Replace[DiagonalMatrix[Energy/@syst]/. 0.->0,x_/;x=!=0->f,{2}]]*)

(*MakeTransformMatrix[system:{_AtomicState..},freq_,rwt_]:=Module[{rwt1},
	rwt1=If[rwt===Automatic,RotatingWaveTransformMatrix[system,freq],rwt];
	Which[OperatorQ[rwt1],rwt1,OperatorQ[rwt[system,freq]],rwt1[system,freq]]
]*)

(*RotatingWaveTransform[syst:{_AtomicState..},H_,f_,t_]:=TrigToExp@EffectiveHamiltonian[H,RotatingWaveTransformMatrix[syst,f],t]*)

Options[RotatingWaveTransform] = {TransformMatrix -> Automatic, TimeVariable -> Automatic, Method -> Automatic}

RotatingWaveTransform[system:{_AtomicState..},op_?OperatorQ,freq_,opts:OptionsPattern[]]:=Module[
	{t = OptionValue[TimeVariable] /. Automatic -> TimeVariable /. Options[DensityMatrix], rwt,
	newopts = FilterRules[Join[{opts},Options[RotatingWaveTransform]],Options[RotatingWaveTransformMatrix]]},
	rwt = If[OptionValue[TransformMatrix] === Automatic,
		RotatingWaveTransformMatrix[system,freq,Sequence@@newopts], 
		OptionValue[TransformMatrix]];
	EffectiveHamiltonian[op, rwt, t]
]

Options[RotatingWaveApproximation] = Options[RotatingWaveTransform]

RotatingWaveApproximation[system:{_AtomicState..}, op_?OperatorQ, 
  freqlist : {{_, _Rule | {__Rule}, ___} ..} | {Except[_Rule],_Rule|{__Rule},___}, 
  opts : OptionsPattern[]] :=
  Module[ {newopts = FilterRules[Join[{opts}, Options[RotatingWaveApproximation]], Options[RotatingWaveTransformMatrix]], 
    t = OptionValue[TimeVariable] /. Automatic -> TimeVariable /. Options[DensityMatrix], 
    tm, rotatingHam, finalreps,  freqlist1 = If[MatchQ[freqlist,{_,_Rule|{__Rule},___}], {freqlist}, freqlist ]},
   (* Get transform matrix and replacement rules *)
    tm = RotatingWaveTransformMatrix[system, freqlist1, ReturnFrequencyReplacements -> True, Sequence@@newopts];
    (* Apply transform matrix to Hamiltonian *)
    rotatingHam = EffectiveHamiltonian[op,tm[[1]],t];
    (* Write untransformed optical frequencies in terms of transformed optical frequencies and difference frequencies *)
    rotatingHam = 
      Fold[ReplaceAllAt[#1, #2[[1]], Join[Tuples@#2[[2]], Tuples@Reverse@#2[[2]]]]&, rotatingHam, tm[[2]]];      
    (* Drop fast oscillating terms *)
    rotatingHam = 
     DropFastTerms[ExpandAll@rotatingHam, freqlist1[[All, 1]], t];
    (* Substitute back for untransformed optical frequencies *)
    rotatingHam = 
      Fold[ReplaceAllAt[#1, #2[[1]], Join[Tuples@#2[[2]], Tuples@Reverse@#2[[2]]]]&, rotatingHam, tm[[3]]];      
    (* Write all frequences in terms of detunings, if specified *)
    finalreps = 
     Cases[freqlist1 /. {x_, rules : {__Rule}, y___} :> {x, rules[[1]], 
         y}, {freq_, label1_ -> label2_, delta_} :> 
       freq -> Energy[SelectState[system, label2]] - 
         Energy[SelectState[system, label1]] + delta];
    rotatingHam /. finalreps
  ]

RotatingWaveApproximation[system:{_AtomicState..},op_?OperatorQ,freq_,opts:OptionsPattern[]]:=
	Module[{newopts = FilterRules[Join[{opts},Options[RotatingWaveApproximation]], Options[RotatingWaveTransform]],
		t = OptionValue[TimeVariable]/.Automatic->TimeVariable/.Options[DensityMatrix]},
		DropFastTerms[RotatingWaveTransform[system,op,freq,Sequence@@newopts],freq,t]
	]

RotatingWaveApproximation[as:{_AtomicState..},reg_Region,op_?OperatorQ,opts:OptionsPattern[]]/;DefinedQ[reg,OpticalParameters]:=
	Module[{ops=OpticalParameters[reg]},
		RotatingWaveApproximation[as,op,(StandardizeField@@ops)[[1,1]],opts]
	]

RotatingWaveApproximation[as:{_AtomicState..},reg_Region,op_?OperatorQ,opts:OptionsPattern[]]/;!DefinedQ[reg,OpticalParameters] := op

RotatingWaveApproximation[as:{_AtomicState..},es:{_Region..},op:{_?OperatorQ..},opts:OptionsPattern[]] :=
  MapThread[RotatingWaveApproximation[as,#1,#2,opts]&,{es,op}]



(* ::Subsection:: *)
(*Relaxation and repopulation*)


(*Relaxation matrix*)

IntrinsicRelaxation[system:{_AtomicState..}]=WignerEckart[system,{NaturalWidth,0,0}]

TransitRelaxation[system:{_AtomicState..},gt_]:=gt WignerEckart[system,{Identity,0,0}]

(*Transit ground-state repopulation matrix*)

TransitRepopulation[system:{_AtomicState..},gt_]:=Module[{g=SublevelMultiplicity[GroundStates[system]]},
	gt WignerEckart[system,{GroundState,0,0}]/If[g==0,1,g]
]



(* ::Subsubsection:: *)
(*Repopulation due to spontaneous decay*)


Options[OpticalRepopulation]=Join[{HyperfineCoherenceTransfer->False,ExpandDipoleRME->True},Options[DMVariables]]

OpticalRepopulation[system:{_AtomicState..},opts___Rule]:=Module[{newopts},
	newopts=Join[DeleteCases[Join[{opts},Options[OpticalRepopulation]],_->Automatic],Options[DensityMatrix]];
	Switch[Representation/.newopts,
		Zeeman,OpticalRepopulationZ[system,Sequence@@newopts],
		PolarizationMoments,OpticalRepopulationPM[system,Sequence@@newopts]
	]
]

OpticalRepopulationZ[system:{_AtomicState..},opts___Rule]:=Module[{dc,br,w,w1,dm,len,hct,dmopts},
	dmopts=FilterRules[{opts},Options[DensityMatrix]];
	(*dc is the cartesian components of the electric dipole operator, transposed to a matrix of vectors *)	
	dc=Transpose[ToCartesian@WignerEckart[system,{Dipole,1}],{3,1,2}];
	dc=If[ExpandDipoleRME/.{opts}/.Options[OpticalRepopulation],ExpandDipoleRME[system,dc],dc];
	br=DirectSelfProduct[Boole[BranchingRatio[#1][#2]=!=0]&,system];
	hct=HyperfineCoherenceTransfer/.{opts}/.Options[OpticalRepopulation];
	w=DirectSelfProduct[Subtract,Energy/@system];
	w1=If[hct,w,DirectSelfProduct[Subtract,Tr[WignerEckart[system,{Energy,0,0}],List]]];
	dm=DensityMatrix[system,Sequence@@dmopts];
	len=Length[system];
	Table[
		If[PossibleZeroQ[w1[[m,n]]],
			Sum[
				If[PossibleZeroQ[w1[[p,q]]]&& br[[m, q]]!=0 ,
					4/3 br[[m,q]] w[[q,m]]^3 (dc[[p,n]].dc[[m,q]])dm[[q,p]],
					0
				],
				{p,len},{q,len}
			],
			0
		],
	{m,len},{n,len}]
]

OpticalRepopulationPM[system:{_AtomicState..},opts___Rule]:=Module[
{levels=RemoveZeemanSublevels[system]},
upperlevels= Select[levels,!GroundStateQ[#]&];
pm=PolarizationMoments[upperlevels];
If[ExpandDipoleRME/.{opts}/.Options[OpticalRepopulation],ExpandDipoleRME[system,#],#]&@Table[
4/3 Sum[If[Abs[F[le1]-F[le2]]<= k<= F[le1]+F[le2],(Energy[le1]-Energy[l1])^3 ReducedME[l1,{Dipole,1},le1]Component[upperlevels,pm,{le1,le2},k]ReducedME[le2,{Dipole,1},l2] (-1)^(F[le1]+F[le2]+k+1) SJS[{k,F[l2],F[l1]},{1,F[le1],F[le2]}],
Transpose[{ConstantArray[0,2k+1]}]],{le1,upperlevels},{le2,upperlevels}],
{l1,levels},{l2,levels},{k,Abs[F[l1]-F[l2]],F[l1]+F[l2]}]
]

Options[OpticalRepopulationOperator]={HyperfineCoherenceTransfer->False,ExpandDipoleRME->True}

OpticalRepopulationOperator[system:{_AtomicState..},opts___Rule]:=Module[{dm,rep,or},
dm=DensityMatrix[system];
rep=Thread[Flatten[dm]->0];
or=OpticalRepopulation[system,opts];
Transpose[Map[(or/.#->1/.rep)&,dm,{2}],{3,4,2,1}]
]

(*The OpticalRepopulation matrix is equal to Map[Tr[#\[CenterDot]DensityMatrix[System]]&,OpticalRepopulationOperator[System],{2}]*)

(*ExperimentalRegion input*)

Relaxation[as:{_AtomicState..},es:{_Region..}]:=
	(
	IntrinsicRelaxation[as]
	+TransitRelaxation[as,TransitRate[Void][#]]
	+TransitRelaxation[as,Sum[TransitRate[l][#],{l,es}]]
	)&/@es

Repopulation[as:{_AtomicState..},es:{_Region..}]:=
	(
	TransitRepopulation[as,Density[#]TransitRate[Void][#]]
	+OpticalRepopulation[as,DMLabel->RegionLabel[#]]
	+Sum[Density[#]/Density[l] TransitRate[l][#]DensityMatrix[as,DMLabel->RegionLabel[l]],{l,es}]
	)&/@es



(* ::Subsection:: *)
(*FluorescenceOperator*)


Options[FluorescenceOperator]={ExpandDipoleRME->True}

FluorescenceOperator[system:{_AtomicState..},freq_,th_,phi_,opts___Rule]:=
Module[{dc,gr,ikk},
	dc=ToCartesian@WignerEckart[system,{Dipole,1}];
	dc=If[ExpandDipoleRME/.{opts}/.Options[OpticalRepopulation],ExpandDipoleRME[system,dc],dc];
	gr=DiagonalMatrix[Boole/@GroundStateQ/@system];
	(*ikk=IdentityMatrix[3]-DirectSelfProduct[CoordinatesToCartesian[{1,th,phi},Spherical]];*)
	ikk=IdentityMatrix[3]-DirectSelfProduct[{Cos[phi] Sin[th], Sin[phi] Sin[th], Cos[th]}];
	freq^4/(2\[Pi]) Sum[dc[[i]].gr.dc[[j]] ikk[[i,j]],{i,3},{j,3}]
]

FluorescenceOperator[system:{_AtomicState..},freq_,opts___Rule]:=
Module[{dc,gr},
	dc=ToCartesian@WignerEckart[system,{Dipole,1}];
	dc=If[ExpandDipoleRME/.{opts}/.Options[OpticalRepopulation],ExpandDipoleRME[system,dc],dc];
	gr=DiagonalMatrix[Boole/@GroundStateQ/@system];
	4/3 freq^4 Sum[dc[[i]].gr.dc[[i]],{i,3}]
]



(* ::Subsection:: *)
(*Liouville equation*)


(*Make the Liouville equation*)

LiouvilleRHS[system:{_AtomicState..},H_?OperatorQ,GMatrix_?OperatorQ,LMatrix_?OperatorQ,opts___Rule]:=
-I Commutator[H,DensityMatrix[system,opts]]-Anticommutator[DensityMatrix[system,opts],GMatrix]/2+LMatrix

Options[LiouvilleEquation]=Options[DMVariables]

LiouvilleEquation[system:{_AtomicState..},H_?OperatorQ,GMatrix_?OperatorQ,LMatrix_?OperatorQ,opts___Rule]:=
Module[{rep,lhs,rhs,newopts,tv,cev,H1,GMatrix1,LMatrix1},
	newopts=DeleteCases[Join[{opts},Options[LiouvilleEquation]],_->Automatic];
	{rep,tv,cev}={Representation,TimeVariable,ComplexExpandVariables}/.newopts/.Options[DensityMatrix];
	lhs=D[DensityMatrix[system,Sequence@@newopts],tv];
	{H1,GMatrix1,LMatrix1}={H,GMatrix,LMatrix}/.ExpandDMVariablesRule[system,Sequence@@newopts];
	rhs=LiouvilleRHS[system,H1,GMatrix1,LMatrix1,Sequence@@newopts];
	If[rep===PolarizationMoments,
		MapThread[Equal,{Simplify@Flatten@Decomposition[system,lhs],Flatten@Decomposition[system,rhs]},1],
			If[cev||cev===Subscript,
				DeleteCases[Flatten@ERI@UpperTriangle[#],True]&,
				Flatten
			]@MapThread[Equal,{lhs,rhs},2]
	]
]

LiouvilleEquation[system:{_AtomicState..},H_?OperatorQ,GMatrix_?OperatorQ,opts___Rule]:=LiouvilleEquation[system,H,GMatrix,0IdentityMatrix[Length[system]],opts]

LiouvilleEquation[system:{_AtomicState..},H_?OperatorQ,opts___Rule]:=LiouvilleEquation[system,H,0IdentityMatrix[Length[system]],0IdentityMatrix[Length[system]],opts]

LiouvilleEquation[
	as:{_AtomicState..},
	es:{_Region..},H:{_?OperatorQ..},
	GMatrix:{_?OperatorQ..},
	LMatrix:{_?OperatorQ..},
	opts:OptionsPattern[]]:=
	Flatten[MapThread[LiouvilleEquation[as,#2,#3,#4,DMLabel->RegionLabel[#1],opts]&,{es,H,GMatrix,LMatrix}],1]

(*ExpandedLiouvilleEquation[system:{_AtomicState..},args__?OperatorQ,opts___Rule]:=SeparateDensityMatrix[system,UpperTriangle@LiouvilleEquation[system,args,opts]]*)

LiouvilleEquation[sys:{_AtomicState..},H_?ContravariantDecompositionMatrixQ,G_?ContravariantDecompositionMatrixQ,L_?ContravariantDecompositionMatrixQ,opts___Rule] :=
    Module[ {newopts = Join[DeleteCases[Join[{opts},Options[LiouvilleEquation]],_->Automatic],Options[DensityMatrix]],levels = RemoveZeemanSublevels[sys],pm,n,f,component},
        pm = PolarizationMoments[levels,opts];
        n = Length[levels];
        f = F/@levels;
        component[op_,{i1_,i2_},k_,q_] :=
            op[[i1,i2,k-Abs[f[[i1]]-f[[i2]]]+1,-q+k+1,1]];
        Flatten@Table[
        Transpose[{Table[
        D[component[pm,{i1,i2},k,q],TimeVariable/.newopts]==
        Sum[-I (-1)^(f[[i1]]+f[[i2]]+k) Sqrt[(2k1+1)(2k2+1)]CG[{k1,q1},{k2,q2},{k,q}]SJS[{k1,k2,k},{f[[i2]],f[[i1]],f[[i3]]}]
        *(
        (component[H,{i1,i3},k1,q1]-I/2 component[G,{i1,i3},k1,q1])component[pm,{i3,i2},k2,q2]
        -component[pm,{i1,i3},k1,q1](component[H,{i3,i2},k2,q2]+I/2 component[G,{i3,i2},k2,q2])
        ),
        {i3,n},{k1,Abs[f[[i1]]-f[[i3]]],f[[i1]]+f[[i3]]},{q1,k1,-k1,-1},{k2,Abs[f[[i2]]-f[[i3]]],f[[i2]]+f[[i3]]},{q2,k2,-k2,-1}]+component[L,{i1,i2},k,q],
        {q,k,-k,-1}]}],
        {i1,n},{i2,n},{k,Abs[f[[i1]]-f[[i2]]],f[[i1]]+f[[i2]]}]
    ]

Options[InitialConditions]=DeleteCases[Options[DMVariables],TimeVariable->_]

InitialConditions[system:{_AtomicState..},InitialDM_?OperatorQ,t0_,opts___Rule]:=Module[{lhs,newopts,cev},
newopts=DeleteCases[Join[{opts},Options[LiouvilleEquation]],_->Automatic];
cev=ComplexExpandVariables/.newopts/.Options[DensityMatrix];
lhs=DensityMatrix[system,TimeVariable->t0,Sequence@@newopts];
If[cev||cev===Subscript,ExprToReIm[UpperTriangle[#],Sequence@@newopts]&,Flatten]@MapThread[Equal,{lhs,InitialDM},2]
]

InitialConditions[
	as:{_AtomicState..},
	es:{_Region..},
	InitialDM:{_?OperatorQ..},t0_,opts___Rule]:=
	Flatten[MapThread[InitialConditions[as,#2,t0,DMLabel->RegionLabel[#1],opts]&,{es,InitialDM}],1]

(*ExpandedInitialConditions[system:{_AtomicState..},InitialDM_?OperatorQ,t0_,opts___Rule]:=SeparateDensityMatrix[system,UpperTriangle@InitialConditions[system,InitialDM,t0,opts]]*)



(* ::Subsection:: *)
(*Solution of wave equation*)


WaveEqnSolution[w_, "AngleEllipticity", {{E0_, cphi_}, {al_, ep_}}, {P1_, P2_, P3_, P4_}] = 
   {{((2*Pi*w)/E0)*(Sin[al]*(Sin[ep]*((-P1)*Cos[cphi] + P2*Sin[cphi]) + 
         Cos[ep]*(P4*Cos[cphi] + P3*Sin[cphi])) + Cos[al]*(Cos[ep]*(P2*Cos[cphi] + P1*Sin[cphi]) + 
         Sin[ep]*(P3*Cos[cphi] - P4*Sin[cphi]))), 
      If[Abs[ep]==Pi/4, Indeterminate,
      	((2*Pi*w)/E0)*Sec[2*ep]*
      	(Cos[al]*(Cos[ep]*(P1*Cos[cphi] - P2*Sin[cphi]) + Sin[ep]*(P4*Cos[cphi] + P3*Sin[cphi])) - 
       	Sin[al]*(Sin[ep]*(P2*Cos[cphi] + P1*Sin[cphi]) + Cos[ep]*((-P3)*Cos[cphi] + P4*Sin[cphi])))]
      }, 
    {If[Abs[ep]==Pi/4, Indeterminate,
    	((2*Pi*w)/E0)*Sec[2*ep]*(Cos[al]*(Sin[ep]*(P1*Cos[cphi] - P2*Sin[cphi]) + 
         Cos[ep]*(P4*Cos[cphi] + P3*Sin[cphi])) - Sin[al]*(Cos[ep]*(P2*Cos[cphi] + P1*Sin[cphi]) + 
         Sin[ep]*((-P3)*Cos[cphi] + P4*Sin[cphi])))], 
      	(-((2*Pi*w)/E0))*
      (Sin[al]*(Cos[ep]*(P1*Cos[cphi] - P2*Sin[cphi]) + Sin[ep]*(P4*Cos[cphi] + P3*Sin[cphi])) + 
       Cos[al]*(Sin[ep]*(P2*Cos[cphi] + P1*Sin[cphi]) + Cos[ep]*((-P3)*Cos[cphi] + P4*Sin[cphi])))}}

WaveEqnSolution[w_, "RatioPhase", {{E0_, cphi_}, {\[Chi]_, ph_}}, {P1_, P2_, P3_, P4_}] := 
   {{((2*Pi*w)/E0)*(Cos[cphi]*(P2*Cos[\[Chi]] + (P4*Cos[ph] + P3*Sin[ph])*Sin[\[Chi]]) + 
       Sin[cphi]*(P1*Cos[\[Chi]] + (P3*Cos[ph] - P4*Sin[ph])*Sin[\[Chi]])), 
     (2*Pi*w*Sec[\[Chi]]*(P1*Cos[cphi] - P2*Sin[cphi]))/E0}, 
    {2*Pi*w*((P4*Cos[ph + cphi]*Cos[\[Chi]] + P3*Cos[\[Chi]]*Sin[ph + cphi] - 
        (P2*Cos[cphi] + P1*Sin[cphi])*Sin[\[Chi]])/E0), If[\[Chi] === 0, 0, 
      (1/E0)*2*Pi*w*(P3*Cos[ph + cphi]*Csc[\[Chi]] - P1*Cos[cphi]*Sec[\[Chi]] + P2*Sec[\[Chi]]*Sin[cphi] - 
        P4*Csc[\[Chi]]*Sin[ph + cphi])]}}

WaveEqnSolution[w_, "Stokes", {{E0_, cphi_}, {S1_, S2_, S3_}}, {P1_, P2_, P3_, P4_}] = 
   {{((Sqrt[2]*Pi*w)/(E0*Sqrt[1 + S1]))*(P2*(1 + S1)*Cos[cphi] + P1*(1 + S1)*Sin[cphi] + 
       Sqrt[1 - S1^2]*(P4*Cos[cphi + ArcTan[S2, S3]] + P3*Sin[cphi + ArcTan[S2, S3]])), 
     (2*Sqrt[2]*Pi*w*(P1*Cos[cphi] - P2*Sin[cphi]))/(E0*Sqrt[1 + S1])}, 
    {(-((2*Sqrt[2]*Pi*w)/E0))*(P2*(-1 + S1)*Sqrt[1 + S1]*Cos[cphi] + P4*Sqrt[1 - S1]*(1 + S1)*
        Cos[cphi + ArcTan[S2, S3]] + P1*(-1 + S1)*Sqrt[1 + S1]*Sin[cphi] + 
       P3*Sqrt[1 - S1]*(1 + S1)*Sin[cphi + ArcTan[S2, S3]]), 
     (-((2*Sqrt[2]*Pi*w)/(E0*(-1 + S1)*Sqrt[1 + S1])))*((-1 + S1)*(P2*S1*S2 - P1*S3)*Cos[cphi] + 
       Sqrt[1 - S1^2]*(P4*S1*S2 - P3*S3)*Cos[cphi + ArcTan[S2, S3]] + (-1 + S1)*(P1*S1*S2 + P2*S3)*
        Sin[cphi] + Sqrt[1 - S1^2]*(P3*S1*S2 + P4*S3)*Sin[cphi + ArcTan[S2, S3]]), 
     (-((2*Sqrt[2]*Pi*w)/(E0*(-1 + S1)*Sqrt[1 + S1])))*((-1 + S1)*(P1*S2 + P2*S1*S3)*Cos[cphi] + 
       Sqrt[1 - S1^2]*(P3*S2 + P4*S1*S3)*Cos[cphi + ArcTan[S2, S3]] + (-1 + S1)*((-P2)*S2 + P1*S1*S3)*
        Sin[cphi] + Sqrt[1 - S1^2]*((-P4)*S2 + P3*S1*S3)*Sin[cphi + ArcTan[S2, S3]])}}



(* ::Subsection:: *)
(*Find dipole polarization*)


(*Lab frame density matrix in terms of the rotating-frame density-matrix elements. Under a passive transformation U, we have  |*)
(*Overscript[\[Psi], ~]\[RightAngleBracket]=U^\[Dagger]|\[Psi]\[RightAngleBracket]. Thus *)
(*Overscript[\[Rho], ~]=|*)
(*Overscript[\[Psi], ~]\[RightAngleBracket]\[LeftAngleBracket]*)
(*Overscript[\[Psi], ~]|=U^\[Dagger]|\[Psi]\[RightAngleBracket]\[LeftAngleBracket]\[Psi]|U=U^\[Dagger] \[Rho] U, so \[Rho]=U *)
(*Overscript[\[Rho], ~](U^\[Dagger]).*)

DMFromRotatingDM[system:{_AtomicState..},RWT_?MatrixQ,opts___Rule]:=
	Module[{newopts=FilterRules[DeleteCases[{opts},_->Automatic],Options[DensityMatrix]]},
		RWT.DensityMatrix[system,Sequence@@newopts].ECT[RWT]
	]

Polarization[system:{_AtomicState..},RWT_,opts___Rule]:=
	Tr[DMFromRotatingDM[system,RWT,opts].#]&/@ToCartesian[WignerEckart[system,{Dipole,1}]]

Options[PolarizationComponents]=Join[Join[DeleteCases[Options[OpticalField],CartesianCoordinates->_],Options[DensityMatrix]]/.(x_->_)->(x->Automatic),{TransformMatrix->Automatic}]



(* ::Input:: *)
(*[args__,opts:OptionsPattern[]]/;2<=Length[{args}]<=3*)


(*Not sure why I enforced complex expanding the variables here. Do I have to?*)

PolarizationComponents[system:{_AtomicState..},args__,opts:OptionsPattern[]]/;2<=Length[{args}]<=3 :=
  Module[ {w,k0,E0,cphi,pol,polparams,e,k,rwt,
  	newopts = Append[DeleteCases[Join[{opts},Options[DensityMatrix]],_->Automatic]/.(ComplexExpandVariables->False)->(ComplexExpandVariables->True),ComplexExpandVariables->True]},
    {{w,k0},{E0,cphi},polparams} = StandardizeField[args,opts];
    {e,k} = {Normalize[OptionValue[PolarizationVector]/.Automatic->PolarizationVector/.Options[OpticalField]],Normalize[OptionValue[PropagationVector]/.Automatic->PropagationVector/.Options[OpticalField]]};
    rwt = If[OptionValue[TransformMatrix] === Automatic, 
    	RotatingWaveTransformMatrix[system, w],
    	OptionValue[TransformMatrix]
    	];
    pol = Polarization[system,rwt,Sequence@@newopts];
    If[ Length[pol]===1, pol = pol[[1]]e ];
    Flatten[{1,-1}Coefficient[Expand@ExpToTrig[pol\[CenterDot]#],{Cos[Expand[w t]],Sin[Expand[w t]]}]&/@{e,Cross[k,e]}]
  ]

Options[Observables]=Join[Join[DeleteCases[Options[OpticalField],CartesianCoordinates->_],Options[DensityMatrix]]/.(x_->_)->(x->Automatic),{ExpandDipoleRME->True,TransformMatrix->Automatic}]

Observables[system:{_AtomicState..},args__,opts:OptionsPattern[]]/;2<=Length[{args}]<=3 :=
  Module[ {w,E0,cphi,pol,e,k,k0,newopts,wes,param = OptionValue[Parameterization]/.Automatic->Parameterization/.Options[OpticalField]},
    {{w,k0},{E0,cphi},pol} = StandardizeField[args,opts];
    {e,k} = {Normalize[OptionValue[PolarizationVector]/.Automatic->PolarizationVector/.Options[OpticalField]],Normalize[OptionValue[PropagationVector]/.Automatic->PropagationVector/.Options[OpticalField]]};
    newopts = FilterRules[DeleteCases[Join[{opts},Options[Observables]],_->Automatic],Options[PolarizationComponents]];
    wes = WaveEqnSolution[w,param,{{E0,cphi},pol},PolarizationComponents[system,args,Sequence@@newopts]];
    If[ OptionValue[ExpandDipoleRME],
      ExpandDipoleRME[system,wes],
      wes
    ]
  ]

Observables[as:{_AtomicState..},es:{_Region..},opts:OptionsPattern[]]:=
	Total[Observables[as,Sequence@@(OpticalParameters[#]),DMLabel->RegionLabel[#],opts]&/@es]



(* ::Subsection:: *)
(*Absorption Functions *)


(*lineshape function*)

Voigt[u_,v_,0]:=v/(\[Pi] (u^2+v^2))

Voigt[u_,0,GD_]:=1/(Sqrt[\[Pi]] GD) E^(-(u^2/GD^2))

Voigt[u_,v_,GD_]:=1/(GD Sqrt[\[Pi] ]) Re[E^-((u+I v)/GD)^2 Erfc[-I((u+I v)/GD)]]

InternalEnergy[s_]:=WignerEckart[s,{Energy,0,0},s]

Options[LinearAbsorption]={ExpandDipoleRME->True}

LinearAbsorption[s1_AtomicState,s2_AtomicState,\[CapitalGamma]D_,w_,dz_,OptionsPattern[]]:=
	If[OptionValue[ExpandDipoleRME],
		ExpandDipoleRME[{s1,s2},#],#]&[
			-((4 \[Pi]^2 (Energy[s2]-Energy[s1]))/SublevelMultiplicity[DeleteParameter[s1,F]])Voigt[ w-(InternalEnergy[s2]-InternalEnergy[s1]),NaturalWidth[s2]/2,\[CapitalGamma]D] dz^2
		]

LinearAbsorption[sys:{_AtomicState..},\[CapitalGamma]D_,w_]:=Module[{dz},
	dz=WignerEckart[sys,{Dipole,1,0}];
	Total[Outer[
		If[M[#1]===M[#2]||AtomicSystemType[sys]===Toy,
		LinearAbsorption[#1,#2,\[CapitalGamma]D,w,Component[sys,dz,{#1,#2}]],
		0
	]&,
	GroundStates[sys],ExcitedStates[sys]],2]
]



(* ::Subsection:: *)
(*Constants and conversions*)


(* ::Subsubsection:: *)
(*Physical constants*)


(*In MKS:*)
(*
AtomicDensityMatrixConstants[SI]={\[Epsilon]0->VacuumPermittivity,\[Mu]0->VacuumPermeability,\[HBar]->PlanckConstant/(2 \[Pi]),c->SpeedOfLight,a0->BohrRadius,e->ElectronCharge,BohrMagneton->Evaluate[Convert[ElectronCharge/(2 ElectronMass),Mega/(Gauss Second)]]}
*)


(* ::Subsubsection:: *)
(*Variable conversions*)


(*Reduced dipole matrix element in terms of natural width and transition wavelength (CGS version) [Sob91, Sec. 9.2.1, Eq. (47)].*)
(*
NaturalWidthFromRME[\[Lambda]_,lowerState_,upperState_,CGS]=NaturalWidth[upperState]->(4w^3)/(3 (c^3) ) 1/(2J[upperState]+1) ReducedME[lowerState,{Dipole,1},upperState]^2/.w->2\[Pi] c/\[Lambda]

ReducedDipoleME[\[Lambda]_,lowerState_,upperState_,CGS]=Simplify[Solve[Equal@@NaturalWidthFromRME[\[Lambda],lowerState,upperState,CGS],ReducedME[lowerState,{Dipole,1},upperState]][[2,1]]]
*)
(*Reduced dipole matrix element in terms of natural width and transition wavelength (MKS version) [Cor77, Sec. 4.5, Eq. (23)]. ||d|| is multiplied by \[HBar] because we measure ||d|| in frequency units. *)



(* ::Input:: *)
(*Convert[d->\[CapitalGamma],SI]=Solve[\[CapitalGamma][b]==w^3/(3 \[Pi] \[Epsilon]0 \[HBar] (c^3) ) 1/(2J[b]+1) (RME[a_,{Dipole,1},b_] \[HBar])^2/.Convert[w->\[Lambda]],RME[a_,{Dipole,1},b_]][[2]]*)


(*Light electric field amplitude in terms of light intensity (MKS version).*)



(* ::Input:: *)
(*Convert[E0->Intensity,SI]=Solve[LightPower/BeamArea==E0[_]^2/(2 c \[Mu]0 ),E0[_]][[2]]*)


(*Here we use the above formula to write the electric field amplitude in terms of light intensity when power is measured in mW and area in cm^2.*)



(* ::Input:: *)
(*E0[12]/.Convert[E0->Intensity,SI]/.PhysicalConstants[SI]/.{LightPower->LightPower (Milli Watt),BeamArea->BeamArea (Centimeter)^2}//Convert[#,Volt/Centimeter]&//N*)


(* ::Subsubsection:: *)
(*Conversion utilities*)


(*Convert the units of each side of each of a list of equations.*)



(* ::Input:: *)
(*ConvertEquationList[eqnList_,units_]:=Map[Convert[Expand[PowerExpand[#]],units]&,eqnList,3]*)


(* ::Subsection:: *)
(*End the private context*)


End[]



(* ::Section:: *)
(*Epilog*)


(*This section protects exported symbols and ends the package.*)



(* ::Subsection:: *)
(*Protect exported symbols*)


(*Protect[Evaluate[$Context<>"*"]]*)



(* ::Subsection:: *)
(*End the package context*)


EndPackage[]

(* Load related packages *) 

Needs["AtomicDensityMatrix`AtomicData`"]

Needs["AtomicDensityMatrix`LevelDiagram`"]

Needs["AtomicDensityMatrix`ADMPerturbativeSolve`"]

Needs["AtomicDensityMatrix`MatrixRSolve`"]
